#include <stdio.h>

#include "core/jobs/jobs.h"
#include "core/jobs/jobsman.h"
#include "core/improc/map2d.h"
#include "core/improc/float.h"
#include "core/improc/meshgrid.h"
#include "core/utils/math.h"

#include "hariharan.h"
#include "rev_polish_job.h"
#include "parallel_job.h"

#include <chrono>

// main entry point
int main(int argc, char** argv)
{
	try
	{
		// configure maximum number of worker threads
		get_instance<thread_pool_manager>()->set_max_workers(12);

		// start job manager
		get_instance<jobs_manager>()->start();

		// consts
		const size_t width = 1024;
		const size_t height = width;

		const float bias = (float)0.1;
		const float modulation = (float)0.7;

		const float diam = (float)6.0e-3;
		const float pxsize = diam / (float)width;
		const float radcurv = (float)100e-3;
		const float wavelength = (float)546e-9;

		// create
		printf("creating synthetic fringes... ");

		// create grid
		auto grid = create_centered_grid<float>(width, height, { pxsize, pxsize }, { 0, 0 });

		// create wavefront
		auto wavefront = -hypotsq(grid.xx, grid.yy) / (2.0 * radcurv);
		auto ini_phase = mod(wavefront * (float)(__2PI / wavelength), (float)__2PI);

		// create fringes images
		std::array<map2d<float>, 5> fringes;

		for (size_t i = 0; i < 5; i++)
			fringes[i] = bias + modulation * 0.5f * (1.0f + cos(ini_phase + (float)__2PI * ((float)i / 4.0f)));

		printf("done\r\n");

#ifdef _DEBUG

		// create hariharan job
		printf("creating job... ");

		auto p = std::make_shared<hariharan>(fringes);

		printf("done\r\n");

		// queue job
		printf("waiting for job... ");

		wait_for({ p });

		printf("done\r\n");

		// print results
		auto res_phase = p->get_phase();
		
		auto diff = res_phase - ini_phase;

		auto ofs = atan2(mean(sin(diff)), mean(cos(diff)));

		diff = mod((float)__PI + diff - ofs, (float)__2PI) - (float)__PI;

		printf("RMS: %.5e\r\n", rms(diff));

#else

		const size_t num_repet = 10;

		// first benchmarking
		printf("\r\n");
		printf("BENCHMARKING 1\r\n");
		printf("==============\r\n");
		printf("\r\n");

		// run once to preallocate memory
		wait_for({ std::make_shared<hariharan>(fringes) });

		printf("#num threads\tavg time (sec)\tstdev time (sec)\r\n");

		for (size_t num_workers = 1; num_workers <= 12; num_workers++)
		{
			// change number of max worker threads
			get_instance<thread_pool_manager>()->set_max_workers(num_workers);
			get_instance<thread_pool_manager>()->reset();

			double sum = 0, sum_sq = 0;

			for (size_t j = 0; j < num_repet; j++)
			{
				map2d<float> res_phase(width, height), res_contrast(width, height);

				size_t y_step = (height + num_workers - 1) / num_workers;

				get_instance<thread_pool_manager>()->set_max_workers(num_workers);

				auto _func = [&](const size_t x, const size_t y)
				{
					auto T1 = fringes[3](x, y) - fringes[1](x, y);
					auto T2 = fringes[0](x, y) + fringes[4](x, y) - 2.0f * fringes[2](x, y);
					auto T3 = fringes[0](x, y) + fringes[1](x, y) + 2.0f * fringes[2](x, y) + fringes[3](x, y) + fringes[4](x, y);

					res_phase(x, y) = std::atan2(-2.0f * T1, -T2);
					res_contrast(x, y) = 3.0f * sqrt(4.0f * T1 * T1 + T2 * T2) / (2.0f * T3);
				};

				// get initial time
				auto tick1 = std::chrono::high_resolution_clock::now();

				// run jobs
				std::list<std::shared_ptr<job_base>> list;

				for (size_t i = 0; i < num_workers; i++)
					list.push_back(std::make_shared<parallel_job>(0, width, i * y_step, std::min((i + 1) * y_step, height), _func));

				wait_for(list);

				// get current time
				auto tick2 = std::chrono::high_resolution_clock::now();

				// compute time
				auto delta_ticks = std::chrono::duration_cast<std::chrono::duration<double>>(tick2 - tick1);

				double value = delta_ticks.count();

				sum += value;
				sum_sq += value * value;
			}

			// display mean and stdev
			double mean_val = sum / (double)num_repet;
			double stdev_val = sqrt(sum_sq / (double)num_repet - mean_val * mean_val);

			printf("%zu\t%.6f\t%.6f\r\n", num_workers, mean_val, stdev_val);
		}

		// second benchmarking
		printf("\r\n");
		printf("BENCHMARKING 2\r\n");
		printf("==============\r\n");
		printf("\r\n");

		// run once to preallocate memory
		wait_for({ std::make_shared<hariharan>(fringes) });

		printf("#num threads\tavg time (sec)\tstdev time (sec)\r\n");

		for (size_t num_workers = 1; num_workers <= 12; num_workers++)
		{
			// change number of max worker threads
			get_instance<thread_pool_manager>()->set_max_workers(num_workers);
			get_instance<thread_pool_manager>()->reset();

			double sum = 0, sum_sq = 0;

			for (size_t j = 0; j < num_repet; j++)
			{
				// get initial time
				auto tick1 = std::chrono::high_resolution_clock::now();

				// process job
				wait_for({ std::make_shared<hariharan>(fringes) });

				// get current time
				auto tick2 = std::chrono::high_resolution_clock::now();

				// compute time
				auto delta_ticks = std::chrono::duration_cast<std::chrono::duration<double>>(tick2 - tick1);

				double value = delta_ticks.count();

				sum += value;
				sum_sq += value * value;
			}

			// display mean and stdev
			double mean_val = sum / (double)num_repet;
			double stdev_val = sqrt(sum_sq / (double)num_repet - mean_val * mean_val);

			printf("%zu\t%.6f\t%.6f\r\n", num_workers, mean_val, stdev_val);
		}

#endif

		// return successfuly
		return 0;
	}
	catch (exception_base& e)
	{
		printf("error: %s\r\n", e.to_string().c_str());

		_critical("%s", e.to_string().c_str());
	}
	catch (std::exception& e)
	{
		printf("error: %s\r\n", e.what());

		_critical("%s", e.what());
	}
	catch (...)
	{
		printf("unknown error!\r\n");

		_critical("unknown error!");
	}

	// signal error
	return 1;	
}